def getBeanName(cls):
    if isinstance(cls, str):
        return cls
    from scopeton.objects import Bean
    if isinstance(cls, Bean):
        return cls.name
    return cls.__name__


def callMethodByName(obj, name, *args, **kwargs):
    if hasattr(obj, name):
        return getattr(obj, name)(*args, **kwargs)


class ScopetonException(Exception):
    def __init__(self, *args: object, **kwargs: object) -> None:
        super().__init__(*args, **kwargs)


from scopeton.scopeTools import getBeanName


class Bean(object):
    name = ""
    cls = None  # type: type
    lazy = False
    singleton = True

    def __init__(self, cls, name=None, lazy = False, singleton=True, service = True):
        self.cls = cls
        self.name = getBeanName(name) if name else getBeanName(cls)
        self.lazy = lazy
        self.singleton = singleton
        self.service = service
        if not singleton and service:
            raise Exception("Error, cannot initialize service as non singleton")






import inspect
import sys

def isPython3():
    return sys.version_info.major >= 3


def isClass(obj):
    if not isPython3():
        import new
        return type(obj) is new.classobj or type(obj) is type
    else:
        import inspect
        return inspect.isclass(obj)

def objectIsInstance(obj):
    if not isPython3():
        import new
        return type(obj) is new.instance or (hasattr(obj, "__class__") and hasattr(obj.__class__, "__name__"))
    else:
        import inspect
        return inspect.isclass(obj)


def getMethodSignature(method):
    if isPython3():
        return inspect.getfullargspec(method)
    else:
        return inspect.getargspec(method)


import inspect
import logging
from threading import RLock

from scopeton import compat
from scopeton.objects import Bean
from scopeton.scopeTools import getBeanName, callMethodByName, ScopetonException


class Scope(object):
    '''this is servicelocator pattern implementation'''
    def __init__(self, lock=False, initMethod="postConstruct", destroyMethod="preDestroy", parent=None):
        self._singletons = {}  # type: dict[str, Bean]
        self._beans = {}       # type: dict[str, Bean]
        self.lock = lock       # type: RLock
        self.initMethod = initMethod
        self.destroyMethod = destroyMethod
        self.parent = parent    #type: Scope

    def getInstance(self, name):
        if self.lock:
            self.lock.acquire()
        try:
            return self._getInstance(getBeanName(name))
        finally:
            if self.lock: self.lock.release()

    def _getInstance(self, name):
        name = getBeanName(name)
        if name in self._singletons:
            return self._singletons[name]

        if name not in self._beans:
            raise Exception("Error, no such bean:" + str(name))

        bean = self._beans[name]
        if len(compat.getMethodSignature(bean.cls.__init__).args) == 2:
            instance = bean.cls(self)
        else:
            instance = bean.cls()

        if bean.singleton:
            self._singletons[name] = instance
        return instance

    def registerBean(self, *args):
        if self.lock:
            self.lock.acquire()
        try:
            for bean in args:
                self._registerBean(bean)
        finally:
            if self.lock: self.lock.release()

    def _registerBean(self, bean):
        """
        :type bean: Bean
        """
        name = getBeanName(bean)
        logging.debug("Registering:" + name)
        if name in self._beans:
            raise ScopetonException("Error, bean with name {} already registered".format(name))
        self._beans[getBeanName(bean)] = bean

    def runServices(self):
        for k in self._beans:
            bean = self._beans[k]
            if bean.service:
                callMethodByName(self.getInstance(bean), self.initMethod)



from distutils.core import setup
setup(
  name = 'scopeton',
  packages = ['scopeton'],
  version = '1.001',
  description = 'Dependency injection mechanism',
  author = 'Anatolii Yakushko',
  author_email = 'shaddyx@gmail.com',
  url = 'https://github.com/shaddyx/scopeton', # use the URL to the github repo
  download_url = 'https://github.com/shaddyx/scopeton/tarball/0.1',
  keywords = ['dependency', 'injection'], # arbitrary keywords
  classifiers = []
)
